from .mail import (
    Email,
    extension,
    EmailMessage,
    Attachment,
    StrangerMail,
    event,
    EMAIL
)

__all__ = [
    'Email',
    'extension',
    'EmailMessage',
    'Attachment',
    'StrangerMail',
    'event',
    'EMAIL'
]
